# Metaverse I/O Functions Module
# 
# This module handles all communctions with the server.
# It also handles connections.
# 
#
# Module Name		- I/O Functions Module
# Current Revision No.	- 0.6
# Current Maintainter 	- Cruise <cruise@drunkenbastards.org>
# Last update 		- Mon Sep 20 22:07:28 EDT 1999
# Sourced By		- Main Module
#
# Copyright (C) 1999 David Gale <cruise@drunkenbastards.org>
# For more information visit http://metaverse.daytona.com/Metaverse/
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
# USA.


#
# Sends pre-formatted commands to the server.
#
proc SendToServer {what} {
	global MV

	if {$MV(sock) == -1} {return}
	if $MV(debug) {puts "-> $what"}
	catch {
		puts $MV(sock) $what
		flush $MV(sock)
	} error
}

#
# Reads text from the Metaverse server.
#
proc ReadFrom {} {
	global MV
	
	gets $MV(sock) input
	if {$input == "NAMEINUSE"} {
		ProcChat $MV(nick) "The nick you have chosen is in use. Please choose another. You can use the setup button to do this or type '/nick YourNick' in the chat windows." 0 1 white
		if $MV(debug) {puts "<- Disconnected!"}
		Disconnect
		return
	}
	if {[eof $MV(sock)] == 1} {
		if $MV(debug) {puts "<- Disconnected!"}
		Disconnect
		return;
	}
	ProcessInput $input
}

#
# Process Command input
#
proc ProcessCommand {what} {
	global MV

	set cmd [string range $what 0 [expr [string first " " $what] -1]]
	set rest [string range $what [expr [string first " " $what] +1] end]
	set parms [split $rest " "]
	if {$cmd == ""} {set cmd $what}
	switch -exact -- [string tolower $cmd] {
		"/connect" {
			if {[lindex $parms 0] == "" || [lindex $parms 1] == ""} {
				ProcChat $MV(nick) "You must provide a servername and a port with the /connect command! (EX: /connect this.example.com 7000)" 0 1 white
				bell
				return
			}
			Disconnect
			set MV(roomhost) [lindex $parms 0]
			set MV(roomport) [lindex $parms 1]
			Connect
		}
		"/aved" {
			AvEd
		}
		"/nick" {
			if {[lindex $parms 0] == "" || [string first " " $what] == -1} {
				ProcChat $MV(nick) "You must provide a nickname! (EX: /nick I'm_A_Dork)" 0 1 white
				bell
				return
			}
			set newnick [CheckThatNick [lindex $parms 0]]
			if {$newnick == ""} {return}
			if $MV(names) {KillName $MV(nick)}
			set MV(nick) $newnick
			set MV($MV(nick).haschat) 0
			set MV($MV(nick).chatque) {}
			set MV($MV(nick).name_x_offset) $MV(anim.x_off)
			set MV($MV(nick).name_y_offset) $MV(anim.y_off)
			if $MV(names) {ShowName $MV(nick)}
			SendToServer "NICK $MV(nick)"
		}
		"/whois" {
			if {[lindex $parms 0] == "" || [string first " " $what] == -1} {
				ProcChat $MV(nick) "You must provide a nickname! (EX: /whois Some_Dork)" 0 1 white
				bell
				return
			}
			foreach person $MV(people) {
				if {[string tolower $person] == [string tolower [lindex $parms 0]]} {
					SendToServer "WHOIS $person"
				}
			}
		}
		"/msg" {
			set rest [string range $rest [expr [string first " " $rest] +1] end]
			if {[lindex $parms 0] == "" || $rest == "" || [string first " " $what] == -1} {
				ProcChat $MV(nick) "You must provide a nickname and message text! (EX: /msg Some_Dork HI!)" 0 1 white
				bell
				return
			}
			foreach person $MV(people) {
				if {[string tolower $person] == [string tolower [lindex $parms 0]]} {
					SendToServer "PRIVMSG $person $rest"
					ProcChat $MV(nick) "$rest" 0 0 "light blue"
				}
			}
		}
	}
}

#
# Connect to the current default server.
#
proc Connect {} {
	global MV

	set MV(sock) [socket $MV(roomhost) $MV(roomport)]
	fconfigure $MV(sock) -blocking 0 -buffering line
	fileevent $MV(sock) readable ReadFrom
	SendToServer "AUTH $MV(nick) $MV(x) $MV(y) $MV(avatar) $MV(anim.x_off) $MV(anim.y_off) [file size "$MV(images)/$MV(avatar)"] $MV(anim.baloon_x) $MV(anim.baloon_y)"
	.menu.r.b.connect configure -text "Disconnect" -command Disconnect
}

#
# Connect directly to a room.
#
proc ConnectToRoom {host port} {
	global MV

	Disconnect
	set MV(roomhost) $host
	set MV(roomport) $port
	set MV(sock) [socket $MV(roomhost) $MV(roomport)]
	fconfigure $MV(sock) -blocking 0
	fileevent $MV(sock) readable ReadFrom
	SendToServer "AUTH $MV(nick) $MV(x) $MV(y) $MV(avatar) $MV(anim.x_off) $MV(anim.y_off) [file size "$MV(images)/$MV(avatar)"] $MV(anim.baloon_x) $MV(anim.baloon_y)"
	.menu.r.b.connect configure -text "Disconnect" -command Disconnect
}

#
# Disconnect from the server.
#
proc Disconnect {} {
	global MV

	if {$MV(sock) == -1} {return}
	catch {close $MV(sock)}
	set MV(sock) -1
	wm title . "*** Not Connected! ***"
	.menu.r.b.connect configure -text "Connect" -command Connect
	foreach who $MV(people) {
		unset MV($who.name)
		unset MV($who.avatar)
		unset MV($who.x)
		unset MV($who.y)
		unset MV($who.lastsaid)
		unset MV($who.name_x_offset)
		unset MV($who.name_y_offset)
		unset MV($who.baloon_x)
		unset MV($who.baloon_y)
		unset MV($who.haschat)
		unset MV($who.nomoremove)
		unset MV($who.chatque)
		set idx [lsearch -exact $MV(people) $who]
		set MV(people) [lreplace $MV(people) $idx $idx]
		.top.c delete $MV($who.icon)
		if $MV(names) {KillName $who}
	}
	DoNames

	#
	# Proces Plugins!
	#
	foreach plugin $MV(plugin.traps.Disconnect) {
		if ![$MV(plugin.traps.Disconnect.$plugin)] {return}
	}
}

#
# Process Server input
#
proc ProcessInput {what} {
	global MV

	if $MV(debug) {puts "<- $what"}
	set cmd [string range $what 0 [expr [string first " " $what] -1]]
	set rest [string range $what [expr [string first " " $what] +1] end]
	set parms [split $rest " "]
	if {$cmd == ""} {set cmd $what}

	switch -exact -- $cmd {
		"NEW" {
			NewPerson [lindex $parms 0] [lindex $parms 1] [lindex $parms 2] [lindex $parms 3] [lindex $parms 4] [lindex $parms 5] [lindex $parms 6] [lindex $parms 7] [lindex $parms 8]
		}
		"NOMORE" {
			PersonLeft [lindex $parms 0]
		}
		"USERS" {
			set MV(server_users) [lindex $parms 0]
		}
		"SEND" {
			SendFileToServer [lindex $parms 0] [lindex $parms 1]
		}
		"MOVE" {
			MoveUser [lindex $parms 0] [lindex $parms 1] [lindex $parms 2] [lindex $parms 3]
		}
		"WHOIS" {
			WHOISUser [lindex $parms 0] [lindex $parms 1]
		}
		"ROOM" {
			SetRoom [lindex $parms 0] [lindex $parms 1]
		}
		"ROOMNAME" {
			wm title . $rest
			set MV(roomname) $rest
		}
		"PING" {
			SendToServer "PONG"
		}
		"EXIT" {
			if !$MV(downloadingroom) {
				set MV(roomhost) [lindex $parms 0]	
				set MV(roomport) [lindex $parms 1]
				Disconnect
				ProcChat $MV(nick) "Changing Rooms to [lindex $parms 0] [lindex $parms 1]" 0 1 white
				Connect	
			}
		}
		"GET" {
			GetAvatar [lindex $parms 0] [lindex $parms 1] [lindex $parms 2]
		}
		"RGET" {
			GetRoom [lindex $parms 0] [lindex $parms 1] [lindex $parms 2]
		}
		"CHAT" {
			set parms [split $rest " "]
			set rest [string range $rest [expr [string first " " $rest] +1] end]
			ProcChat [lindex $parms 0] $rest 0 0 white
		}
		"PRIVMSG" {
			set parms [split $rest " "]
			set rest [string range $rest [expr [string first " " $rest] +1] end]
			ProcChat [lindex $parms 0] "$rest" 0 0 "light blue"
		}
		"AVATAR" {
			ChangeUserAvatar [lindex $parms 0] [lindex $parms 1] [lindex $parms 2] [lindex $parms 3] [lindex $parms 4] [lindex $parms 5] [lindex $parms 6]
		}
	}
}

proc SendFileToServer {port what} {
	global MV

	if ![SanityCheck "$what"] {return}
	if ![file exists "$MV(images)/$what"] {return}
	set size [file size "$MV(images)/$what"]


	set rc [DownloadWindow $what]
	set row [string range $rc 0 [expr [string first " " $rc] -1]]
	set col [string range $rc [expr [string first " " $rc] +1] end]

	set MV(dlbytes.$row.$col) 0
	set MV(dlperc.$row.$col) 0

	set infile [open "$MV(images)/$what" RDONLY]
	fconfigure $infile -translation binary
	set sock [socket $MV(roomhost) $port]
	fconfigure $sock -blocking 0 -buffering none -translation binary
	while {[eof $infile] != 1} {
		set buffer [read $infile $MV(sendbuffer)]
		puts -nonewline $sock $buffer
		incr MV(dlbytes.$row.$col) [string length $buffer]
		if {$MV(dlbytes.$row.$col) > 0} {
			set MV(dlperc.$row.$col) [expr ($MV(dlbytes.$row.$col) * 100) / $size]
			ShowDownloadStatus $row $col
			if $MV(debug) {puts "Send Buffer [string length $buffer] $sock"}
		}
		update idletasks
		update
		after 70
	}
	close $infile
	close $sock
	KillDlwin $row $col
}
proc SanityCheck {what} {
        if {[string first "../" $what] != -1} { return 0} else { return 1}
}

proc GetAvatar {port what size} {
	global MV


	if ![SanityCheck "$what"] {return}
	if {[lsearch $MV(downloads) $what] != -1} {return}
	if [file exists "$MV(rem_images)/$what"] {return}
	lappend MV(downloads) $what
	set sock -1
	set sock [socket $MV(roomhost) $port]
	if {$sock == -1} {
		if $MV(debug) {puts "Failed to open connection!\n"}
		set idx [lsearch -exact $MV(downloads) $what]
		set MV(downloads) [lreplace $MV(downloads) $idx $idx]
		return
	}

	set rc [DownloadWindow $what]
	set row [string range $rc 0 [expr [string first " " $rc] -1]]
	set col [string range $rc [expr [string first " " $rc] +1] end]

	set MV(dlbytes.$row.$col) 0
	set MV(dlperc.$row.$col) 0
	set MV(dlname.$row.$col) $what

	set flags [list WRONLY CREAT]
	set outfile [open "$MV(rem_images)/$what" $flags 0600]
	fconfigure $outfile -translation binary
	fconfigure $sock -buffering none -blocking 0 -translation binary -buffersize $MV(sendbuffer)

	fileevent $sock readable "ReadForAvatar $sock $outfile $row $col  $size"
	update idletasks
}

proc ReadForAvatar {sock outfile row col size} {
	global MV

	set what $MV(dlname.$row.$col)
	set buffer [read $sock $MV(getbuffer)]
	incr MV(dlbytes.$row.$col) [string length $buffer]
	if {$MV(dlbytes.$row.$col) > 0} {
		set MV(dlperc.$row.$col) [expr ($MV(dlbytes.$row.$col) * 100) / $size]
		ShowDownloadStatus $row $col
		if $MV(debug) {puts "Read Buffer [string length $buffer] $sock"}
		puts -nonewline $outfile $buffer
		flush $outfile
	}
	if {[eof $sock] == 1} {
		close $outfile
		close $sock	
		KillDlwin $row $col
		set idx [lsearch -exact $MV(downloads) $what]
		set MV(downloads) [lreplace $MV(downloads) $idx $idx]
		set sze 0
		update idletasks
		catch {set sze [file size "$MV(rem_images)/$what"]}
		if {$sze != $size} {
			if $MV(debug) {puts "ERROR, Size Mismatch local=$sze remote=$size"}
			catch {file delete "$MV(rem_images)/$what"}
			if {[lsearch -exact $MV(downloads) $what] == -1} {
				SendToServer "SEND $what"
			}
			update idletasks
			return
		}
		foreach who $MV(people) {
			if {$MV($who.avatar) == $what} {
				catch {image create photo $who -file "$MV(rem_images)/$what"}
			}
		}
	}
	update idletasks
}

proc GetRoom {port what size} {
	global MV
	

	if ![SanityCheck "$what"] {return}

	set rc [DownloadWindow $what]
	set row [string range $rc 0 [expr [string first " " $rc] -1]]
	set col [string range $rc [expr [string first " " $rc] +1] end]

        set MV(dlbytes.$row.$col) 0
	set MV(dlperc.$row.$col) 0

        update idletasks


	if [file exists "$MV(roomdir)/$what"] {return}
	lappend MV(roomloads) $what
	set MV(downloadingroom) 1
	set flags [list WRONLY CREAT]
	set outfile [open "$MV(roomdir)/$what" $flags 0600]
	fconfigure $outfile -translation binary
	set sock [socket $MV(roomhost) $port]
	fconfigure $sock -buffering none -blocking 0 -translation binary -buffersize $MV(getbuffer)
	while {[eof $sock] != 1} {
		set buffer [read $sock $MV(getbuffer)]
		incr MV(dlbytes.$row.$col) [string length $buffer]
		set MV(dlperc.$row.$col) [expr ($MV(dlbytes.$row.$col) * 100) / $size]
		ShowDownloadStatus $row $col
		if $MV(debug) {puts "Read Buffer [string length $buffer] $sock"}
		puts -nonewline $outfile $buffer
		update idletasks
		update 
	}
	close $outfile
	close $sock	

	KillDlwin $row $col
	if {[file size "$MV(roomdir)/$what"] != $size} {
		catch {file delete "$MV(roomdir)/$what"}
		SendToServer "RSEND $what"
		return
	}	
	set idx [lsearch -exact $MV(roomloads) $what]
	set MV(roomloads) [lreplace $MV(roomloads) $idx $idx]
	image create photo room -file "$MV(roomdir)/$what"
	set MV(downloadingroom) 0
}

proc DownloadWindow {file} {
	global MV
	
	if {[string length $file] > 16} {
		set file [string range $file 0 15]
	}

	# find an empty one.
	set row -1
	set col -1
	for {set c 0} {$c < 26 && $row < 0} {incr c} {
		for {set s 1} {$s < 4 && $col < 0} {incr s} {
			if !$MV(dls.$c.$s) {
				set row $c
				set col $s
			}
		}
	}
	# Make it not empty :)
	set y [expr $row * 18 + 2]
	set xx [expr $col * 213]
	set MV(dls.$row.$col) 1
	set MV(dlbox.$row.$col) [.top.c create rectangle [expr $xx - 213] [expr $y +1] $xx [expr $y + 18 + 2] -fill black -outline white]
 	set MV(dltxt.$row.$col) [.top.c create text [expr $xx - 140] [expr $y + 10] -text "$file" -fill white]
	set MV(dlstsbx.$row.$col) [.top.c create rectangle [expr $xx -68] [expr $y + 4] [expr $xx - 2] [expr $y + 16] -outline blue]
	set MV(dlsts.$row.$col) [.top.c create rectangle [expr $xx -66] [expr $y + 5] [expr $xx -66] [expr $y + 14] -fill red]
	set MV(dlperc.$row.$col) 0
	return "$row $col"
}

proc ShowDownloadStatus {row col} {
	global MV

	set y [expr $row * 18 + 2]
	set xx [expr $col * 213]
	set rightx [expr ($MV(dlperc.$row.$col) * 62) / 100]
	set leftx [expr $xx -66]

	.top.c delete $MV(dlsts.$row.$col)
	set MV(dlsts.$row.$col) [.top.c create rectangle $leftx [expr $y + 5] [expr $leftx + $rightx] [expr $y + 14] -fill red]
}

proc KillDlwin {row col} {
	global MV

	.top.c delete $MV(dlbox.$row.$col)
	.top.c delete $MV(dltxt.$row.$col)
	.top.c delete $MV(dlstsbx.$row.$col)
	.top.c delete $MV(dlsts.$row.$col)
	set MV(dls.$row.$col) 0
}
