#!/usr/bin/tcl
# Metaverse Server Program
# 
# This is the server code! :)
#
# Module Name		- Server Program
# Current Revision No.	- 0.1
# Current Maintainter 	- Cruise <cruise@drunkenbastards.org>
# Last update 		- Sun Sep 12 13:34:17 EDT 1999
# Sourced By		- Command Line

global MV
global waiter

set MV(socks) {}

set MV(ports) {}
set MV(port) "7000"
set MV(timeout) 120
set app "$argv0"
catch { set app [file readlink $app] }
set MV(homedir) "[file dirname $app]"
set MV(avatars) "$MV(homedir)/images"
set MV(objects) "$MV(homedir)/objects"
set MV(roomdir) "$MV(homedir)/rooms"
set MV(configfile) "$MV(homedir)/$argv"
set MV(roomfile) "hippos.gif"
set MV(sendbuffer) 4096
set MV(exits) {}
set MV(downloads) {}
set MV(locations) {}

if {$argv == ""} {
	puts "Usage: server.tcl <Config File>"
	exit
}
if [file exists $MV(configfile)] {source $MV(configfile)}

socket -server NewConnect $MV(port)

proc bgerror {stuff} {
	global errorInfo
	puts $stuff
	puts $errorInfo
}

proc NewConnect {sck address port} {
	global MV
	puts "($sck)<- New Connection! $address:$port"
	fconfigure $sck -blocking 0 -buffering line
	fileevent $sck readable "ReadFrom $sck"
	lappend MV(socks) $sck
	set MV($sck.name) "*"
	set MV($sck.address) "$address"
	set MV($sck.port) "$port"
	set MV($sck.ping) 0
	set MV($sck.ping_response) [clock seconds]
	set MV($sck.x) -1
	set MV($sck.y) -1
	set MV($sck.avatar) "*connecting*"
	set MV($sck.av_head_x) "-1"
	set MV($sck.av_head_y) "-1"
}

proc SanityCheck {what} {
	if {[string first "../" $what] != -1} { return 0} else { return 1}
}

proc TestNum {what} {
	global MV

	set bad 0
	for {set c 0} {$c < [string length $what] && !$bad} {incr c} {
		switch -exact -- [string range $what $c $c] {
			1 {
			}
			2 {
			}
			3 {
			}
			4 {
			}
			5 {
			}
			6 {
			}
			7 {
			}
			8 {
			}
			9 {
			}
			0 {
			}
			- {
			}
			default {
				set bad 1
			}
		}
	}
	return $bad
}


proc ReadFrom {what} {
	global MV
	
	catch {gets $what input}
	if {[eof $what] == 1} {
		if {$MV($what.name) == "*"} {
			DisconnectUser $what 0
		} else {
			DisconnectUser $what 1
		}
		return
	}

	if {$MV($what.name) == "*"} {
		if {[string range $input 0 3] != "AUTH"} {
			if {$input == "USERS"} {
				SendToUser $what "USERS [expr [llength $MV(socks)] -1]"
				DisconnectUser $what 0
			} else {
				SendToUser $what "AUTH REQD"
				DisconnectUser $what 0
			}
		} else {
			set parms [split [string range $input 5 end] " "]
			if {[TestNum [lindex $parms 1]] || \
			  [TestNum [lindex $parms 2]] || \
			  [TestNum [lindex $parms 4]] || \
			  [TestNum [lindex $parms 5]] || \
			  [TestNum [lindex $parms 6]] || \
			  [TestNum [lindex $parms 7]] || \
			  [TestNum [lindex $parms 8]] } {
				SendToUser $what "AUTH FAILED"
				DisconnectUser $what 0
				return
			}
			if {[string length [lindex $parms 1]] > 3 || \
				[string length [lindex $parms 2]] > 3 || \
				[string length [lindex $parms 4]] > 4 || \
				[string length [lindex $parms 5]] > 4 || \
				[string length [lindex $parms 6]] > 6 || \
				[string length [lindex $parms 7]] > 4 || \
				[string length [lindex $parms 8]] > 4 || \
				[lindex $parms 1] < 0 || \
				[lindex $parms 2] < 0 || \
				[lindex $parms 6] < 0} {
				SendToUser $what "AUTH FAILED"
				DisconnectUser $what 0
				return
			}
			set MV($what.downloads) {}
			set MV($what.name) [string range [lindex $parms 0] 0 12]
			set MV($what.x) [lindex $parms 1]
			set MV($what.y) [lindex $parms 2]
			set MV($what.avatar) [lindex $parms 3]
			set MV($what.av_head_x) [lindex $parms 4]
			set MV($what.av_head_y) [lindex $parms 5]
			set MV($what.av_baloon_x) [lindex $parms 7]
			set MV($what.av_baloon_y) [lindex $parms 8]
			
			set size [lindex $parms 6]
			update idletasks
			if {[string trim $MV($what.name)] == ""} {
					SendToUser $what "NAMEINUSE"
					DisconnectUser $what 0
					return
			}
			foreach sckt $MV(socks) {
				if {$MV($sckt.name) == $MV($what.name) && $sckt != $what} {
					SendToUser $what "NAMEINUSE"
					DisconnectUser $what 0
					return
				}
			}
			SendToUser $what "ROOMNAME $MV(roomname)"
			SendToUser $what "ROOM $MV(roomfile) [file size $MV(roomdir)/$MV(roomfile)]"
			if ![file exists $MV(avatars)/$MV($what.avatar)] {
				SendToAllUsers "NEW $MV($what.name) $MV($what.x) $MV($what.y) default.gif 0 20 [file size $MV(avatars)/default.gif] 24 6"
				puts "($what) $MV($what.avatar) does not exist"
				GetBinaryFile $what $MV($what.avatar)
			} else {
				if {[file size $MV(avatars)/$MV($what.avatar)] != $size} {
					SendToAllUsers "NEW $MV($what.name) $MV($what.x) $MV($what.y) default.gif 0 20 [file size $MV(avatars)/default.gif] 24 6"
					puts "($what) $MV($what.avatar) $size != [file size $MV(avatars)/$MV($what.avatar)]"
					GetBinaryFile $what $MV($what.avatar)
				} else {
					SendToAllUsers "NEW $MV($what.name) $MV($what.x) $MV($what.y) $MV($what.avatar) $MV($what.av_head_x) $MV($what.av_head_y) [file size $MV(avatars)/$MV($what.avatar)] $MV($what.av_baloon_x) $MV($what.av_baloon_y)"
				}
			}
			foreach sckt $MV(socks) {
				if [file exists $MV(avatars)/$MV($sckt.avatar)] {
					SendToUser $what "NEW $MV($sckt.name) $MV($sckt.x) $MV($sckt.y) $MV($sckt.avatar) $MV($sckt.av_head_x) $MV($sckt.av_head_y) [file size $MV(avatars)/$MV($sckt.avatar)] $MV($what.av_baloon_x) $MV($what.av_baloon_y)"
				} else {
					SendToUser $what "NEW $MV($sckt.name) $MV($sckt.x) $MV($sckt.y) default.gif 0 20 [file size $MV(avatars)/default.gif] 24 6"
				}
			}
			return
		}
	}
	ProcessInput $what $input
}

proc DisconnectUser {who announce} {
	global MV
	
	puts "($who)<- Disconnected! $MV($who.address):$MV($who.port)"
	catch {close $who}
	set which [lsearch -exact $MV(socks) $who]
	if {$which == -1} {return}
	set MV(socks) [lreplace $MV(socks) $which $which]
	if $announce {
		foreach sckt $MV(socks) {
			SendToUser $sckt "NOMORE $MV($who.name)"
		}
	}
}

proc SendToUser {who what} {
	puts "($who)-> $what"
	catch {puts $who "$what"}
	#flush $who
}

proc SendToAllUsers {what} {
	global MV
	
	foreach sckt $MV(socks) {
		if {$MV($sckt.name) != "*"} {
			puts "($sckt)-> $what"
			catch {puts $sckt "$what"}
			#flush $sckt
		}
	}
}

proc ProcessInput {who what} {
	global MV

	puts "($who)<- $what"
	set cmd [string range $what 0 [expr [string first " " $what] -1]]
	set rest [string range $what [expr [string first " " $what] +1] end]
	set parms [split $rest " "]
	if {$cmd == ""} {set cmd $what}
	switch -exact -- $cmd {
		"MOVE" {
				#fixes a bug that MbM found.
			  if {[TestNum [lindex $parms 1]] || \
			  [TestNum [lindex $parms 2]] || \
			  [TestNum [lindex $parms 3]]} {return}
			if {[string length [lindex $parms 1]] > 4} {return}
			if {[string length [lindex $parms 2]] > 4} {return}
			if {[string length [lindex $parms 3]] > 2} {return}
			if {[lindex $parms 1] < 0} {return}
			if {[lindex $parms 2] < 0} {return}
			if {[lindex $parms 3] < 0} {return}
			set MV($who.x) [lindex $parms 1]
			set MV($who.y) [lindex $parms 2]
			set is_exiting 0
			set idx 0
			foreach exit $MV(exits) {
				set exl [split $exit " "]
				set x1 	[lindex $exl 0]
				set y1 	[lindex $exl 1]
				set x2 	[lindex $exl 2]
				set y2 	[lindex $exl 3]
				if {$MV($who.x) > $x1 && \
					$MV($who.x) < $x2 && \
					$MV($who.y) > $y1 && \
					$MV($who.y) < $y2} {
						set is_exiting 1
						set eidx $idx
				}
				incr idx
			}
			if $is_exiting {				
				SendToAllUsers "MOVE $MV($who.name) [lindex $parms 1] [lindex $parms 2] [lindex $parms 3]"
				SendToUser $who "EXIT [lindex $MV(locations) $eidx]"
			} else {
				SendToAllUsers "MOVE $MV($who.name) [lindex $parms 1] [lindex $parms 2] [lindex $parms 3]"
			}
		}
		"SEND" {
			if ![SanityCheck [lindex $parms 0]] {return}
			SendBinaryFile $who [lindex $parms 0]
		}
		"USERS"
		{
			SendToUser $what "USERS [llength $MV(socks)]"
		}
		"PONG" {
			set MV($who.ping_response) [clock seconds]
		}
		"RSEND" {
			if ![SanityCheck [lindex $parms 0]] {return}
			SendRoomFile $who [lindex $parms 0]
		}
		"CHAT" {
			if {$rest != ""} {
				SendToAllUsers "CHAT $MV($who.name) [string range $rest 0 256]"
			}
		}
		"AVATAR" {
			  if {[TestNum [lindex $parms 1]] || \
			  [TestNum [lindex $parms 2]] || \
			  [TestNum [lindex $parms 3]] || \
			  [TestNum [lindex $parms 4]] || \
			  [TestNum [lindex $parms 5]]} {
				SendToUser $who "BAD AVATAR"
				return
			}
			if {[string length [lindex $parms 1]] > 4} {return}
			if {[string length [lindex $parms 2]] > 4} {return}
			if {[string length [lindex $parms 3]] > 6} {return}
			if {[string length [lindex $parms 4]] > 4} {return}
			if {[string length [lindex $parms 5]] > 4} {return}
			if {[lindex $parms 3] < 0} {return}
			ChangeAvatar $who [lindex $parms 0] [lindex $parms 1] [lindex $parms 2] [lindex $parms 3] [lindex $parms 4] [lindex $parms 5]
		}
		"WHOIS" {
			if {[lindex $parms 0] == ""} {return}
			foreach s $MV(socks) {
				if {$MV($s.name) == [lindex $parms 0] || [lindex $parms 0] == "*"} {
					SendToUser $who "WHOIS $MV($s.name) $MV($s.name)@$MV($s.address)"
				}
			}
		}
		"PRIVMSG" {
			if {[lindex $parms 0] == ""} {return}
			set stuff [string range $rest [expr [string first " " $rest] +1] end]
			foreach s $MV(socks) {
				if {$MV($s.name) == [lindex $parms 0] || [lindex $parms 0] == "*"} {
					SendToUser $s "PRIVMSG $MV($who.name) $stuff"
				}
			}
		}
		"NICK" {
			if {$MV($who.name) == [lindex $parms 0] || $MV($who.name) == "*"} {
				return
			}
			foreach sckt $MV(socks) {
				if {$MV($sckt.name) == [lindex $parms 0]} {
					SendToUser $who "NAMEINUSE"
				}
			}
			SendToAllUsers "NOMORE $MV($who.name)"
			SendToAllUsers "NEW [lindex $parms 0] $MV($who.x) $MV($who.y) default.gif 0 20 [file size $MV(avatars)/default.gif] 24 6"
 			SendToAllUsers "AVATAR [lindex $parms 0] $MV($who.avatar) $MV($who.av_head_x) $MV($who.av_head_y) [file size $MV(avatars)/$MV($who.avatar)] $MV($who.av_baloon_x) $MV($who.av_baloon_y)"
			set MV($who.name) [lindex $parms 0]
		}
	}
}

proc ChangeAvatar {who what x y size bx by} {
	global MV

	set MV($who.avatar) $what
	set MV($who.av_head_x) $x
	set MV($who.av_head_y) $y
	set MV($who.av_baloon_x) $bx
	set MV($who.av_baloon_y) $by
	update idletasks
	if ![file exists $MV(avatars)/$what] {
		puts "-> $what does not exist"
		GetBinaryFile $who $MV($who.avatar)
	} else {
		if {[file size $MV(avatars)/$what] != $size} {
			puts "-> $what $size != [file size $MV(avatars)/$what]"
			GetBinaryFile $who $MV($who.avatar)
		} else {
			SendToAllUsers "AVATAR $MV($who.name) $what $MV($who.av_head_x) $MV($who.av_head_y) [file size $MV(avatars)/$what] $bx $by"
		}
	}
}

proc SendBinaryFile {who what} {
	global MV
	
	if {[lsearch -exact $MV($who.downloads) $what] != -1} {
		puts "($who) Already getting $what, rejecting."
		return
	}
	if {[lsearch -exact $MV(downloads) $what] != -1} {
		puts "($who) Still downloading $what, rejecting."
		return
	}
	lappend MV($who.downloads) $what
	set bad 1
	while {$bad == 1} {
		set port [expr int((rand() * 4096)) + 10000]
		if {[lsearch -exact $MV(ports) $port] == -1} {
			lappend MV(ports) $port
			set bad 0
		}
	}
	set MV($port.server) [socket -server "SendFile $port" $port]
	set MV($port.remote) $MV($who.address)
	set MV($port.file) $what
	set MV($port.sckt) $who
	set MV($port.name) $MV($who.name)
	set MV($port.time) [clock seconds]
	SendToUser $who "GET $port $what [file size $MV(avatars)/$what]"
}	

proc SendRoomFile {who what} {
	global MV
	
	set bad 1
	while {$bad == 1} {
		set port [expr int((rand() * 4096)) + 10000]
		if {[lsearch -exact $MV(ports) $port] == -1} {
			lappend MV(ports) $port
			set bad 0
		}
	}
	set MV($port.server) [socket -server "SendRoom $port" $port]
	set MV($port.remote) $MV($who.address)
	set MV($port.file) $what
	set MV($port.sckt) $who
	set MV($port.name) $MV($who.name)
	set MV($port.time) [clock seconds]
	SendToUser $who "RGET $port $what [file size $MV(roomdir)/$MV(roomfile)]"
}	

proc GetBinaryFile {who what} {
	global MV
	
	if {[lsearch -exact $MV(downloads) $what] != -1} {return}
	lappend MV(downloads) $what
	set bad 1
	while {$bad == 1} {
		set port [expr int((rand() * 4096)) + 10000]
		if {[lsearch -exact $MV(ports) $port] == -1} {
			lappend MV(ports) $port
			set bad 0
		}
	}
	set MV($port.remote) $MV($who.address)
	set MV($port.file) $what
	set MV($port.name) $MV($who.name)
	set MV($port.av_head_x) $MV($who.av_head_x)
	set MV($port.av_head_y) $MV($who.av_head_y)
	set MV($port.av_baloon_x) $MV($who.av_baloon_x)
	set MV($port.av_baloon_y) $MV($who.av_baloon_y)
	set MV($port.time) [clock seconds]
	set MV($port.server) [socket -server "GetFile $port" $port]
	SendToUser $who "SEND $port $what"
}	

proc GetFile {which sock address their_port} {
	global MV

	catch {close $MV($which.server)}
	set $MV($which.time) 0
	if {$MV($which.remote) != $address} {
		catch {close $sock}
		return
	}
	set MV($which.socket) $sock
	set MV($which.fileopen) 0
	set MV($which.idletime) [clock seconds]
	fconfigure $sock -buffering none -blocking 0 -translation binary -buffersize 4096
	fileevent $sock readable "StoreFile $which"
	##FIXME (timeout!)
}

proc SendFile {which sock address their_port} {
	global MV

	catch {close $MV($which.server)}
	set $MV($which.time) 0
	if {$MV($which.remote) != $address} {
		catch {close $sock}
		set idx [lsearch $MV($MV($which.sckt).downloads) $MV($which.file)]
		set $MV($MV($which.sckt).downloads) [lreplace $MV($MV($which.sckt).downloads) $idx $idx] 
		return
	}
	set MV($which.socket) $sock
	set MV($which.fileopen) 0
	fconfigure $sock -blocking 0 -buffering none -translation binary
	if ![file exists $MV(avatars)/$MV($which.file)] {
		puts "-> File Transfer incomplete $MV($which.file)"
		set idx [lsearch $MV($MV($which.sckt).downloads) $MV($which.file)]
		set $MV($MV($which.sckt).downloads) [lreplace $MV($MV($which.sckt).downloads) $idx $idx] 
		set idx [lsearch -exact $MV(ports) $which]
		set MV(ports) [lreplace $MV(ports) $idx $idx]
		return
	}
	set infile [open $MV(avatars)/$MV($which.file)]
	fconfigure $infile -translation binary
	fconfigure $sock -blocking 0 -buffering none -translation binary
	while {[eof $infile] != 1} {
		puts "Send Block"
		set MV($which.idletime) [clock seconds]
		set buffer [read $infile $MV(sendbuffer)]
		puts -nonewline $sock $buffer
		update idletasks
	}
	close $infile
	flush $sock
	catch {close $sock}
	puts "-> File Transfer Complete $MV($which.file)"
	set idx [lsearch $MV($MV($which.sckt).downloads) $MV($which.file)]
	puts "Removing idx $idx (of [llength $MV($MV($which.sckt).downloads)])"
	set MV($MV($which.sckt).downloads) [lreplace $MV($MV($which.sckt).downloads) $idx $idx] 
	set idx [lsearch -exact $MV(ports) $which]
	set MV(ports) [lreplace $MV(ports) $idx $idx]
}

proc SendRoom {which sock address their_port} {
	global MV

	catch {close $MV($which.server)}
	set $MV($which.time) 0
	if {$MV($which.remote) != $address} {
		catch {close $sock}
		return
	}
	set MV($which.socket) $sock
	set MV($which.fileopen) 0
	fconfigure $sock -blocking 0 -buffering none -translation binary
	if ![file exists $MV(roomdir)/$MV($which.file)] {
		puts "-> File Transfer incomplete $MV($which.file)"
		set idx [lsearch -exact $MV(ports) $which]
		set MV(ports) [lreplace $MV(ports) $idx $idx]
		return
	}
	set infile [open $MV(roomdir)/$MV($which.file)]
	fconfigure $infile -translation binary
	fconfigure $sock -blocking 0 -buffering none -translation binary
	while {[eof $infile] != 1} {
		puts "Send Block"
		set MV($which.idletime) [clock seconds]
		set buffer [read $infile $MV(sendbuffer)]
		puts -nonewline $sock $buffer
		update idletasks
	}
	close $infile
	catch {close $sock}
	puts "-> File Transfer Complete $MV($which.file)"
	set idx [lsearch -exact $MV(ports) $which]
	set MV(ports) [lreplace $MV(ports) $idx $idx]
}

proc StoreFile {who} {
	global MV

	set input [read $MV($who.socket) 4096]
	if {[eof $MV($who.socket)]} {
		catch {close $MV($who.socket)}
		puts "-> File Transfer Complete $MV($who.file)"
		set idx [lsearch -exact $MV(downloads) $MV($who.file)]
		set MV(downloads) [lreplace $MV(downloads) $idx $idx] 
		set idx [lsearch -exact $MV(ports) $who]
		set MV(ports) [lreplace $MV(ports) $idx $idx]
		SendToAllUsers "AVATAR $MV($who.name) $MV($who.file) $MV($who.av_head_x) $MV($who.av_head_y) [file size $MV(avatars)/$MV($who.file)] $MV($who.av_baloon_x) $MV($who.av_baloon_y)"
		if {$MV($who.fileopen)} {
			catch {close $MV($who.fd)}
			set MV($who.fileopen) 0
		}
	}
	set MV($who.idletime) [clock seconds]

	if $MV($who.fileopen) {
		puts -nonewline $MV($who.fd) $input
		catch {flush $MV($who.fd)}
	} else {
		set flags [list WRONLY CREAT]
		set MV($who.fd) [open $MV(avatars)/$MV($who.file) $flags 0600]
		fconfigure $MV($who.fd) -translation binary
		set MV($who.fileopen) 1
		puts -nonewline $MV($who.fd) $input
		catch {flush $MV($who.fd)}
	}
}

proc CheckTimeouts {} {
	global MV

	set tme [clock seconds]
	foreach port $MV(ports) {
		puts "-> Checking TimeOuts ($port)"
		if $MV($port.time) {
			if {[expr $tme - $MV($port.time)] > $MV(timeout)} {
				catch {close $MV($port.server)}
				puts "-> Connection Timed Out $MV($port.file)"
				set idx [lsearch -exact $MV(ports) $port]
				set MV(ports) [lreplace $MV(ports) $idx $idx]
			}
		} else {
			if {[expr $tme - $MV($port.idletime)] > $MV(timeout)} {
				catch {close $MV($port.socket)}
				puts "-> File Transfer Timed Out $MV($port.file)"
				set idx [lsearch -exact $MV(ports) $port]
				set MV(ports) [lreplace $MV(ports) $idx $idx]
				if {$MV($port.fileopen)} {
					catch {close $MV($port.fd)}
					set MV($port.fileopen) 0
				}
				file remove $MV(avatars)/$MV($port.file)
			}
		}
	}
	set disco {}
	foreach who $MV(socks) {
		if {[expr $tme - $MV($who.ping_response)] > 320} {
			puts "($who) - Ping Timeout!"
			lappend disco $who
		} else {
			if {[expr $tme - $MV($who.ping)] > 150} {
				set MV($who.ping) [clock seconds]
				SendToUser $who "PING"
			}
		}
	}
	foreach who $disco {
		DisconnectUser $who 1
	}
	after 5000 CheckTimeouts
}

CheckTimeouts
vwait waiter
